-----------------------------------------------------------------------------
-- RapidCast - casting bar lag adjustment addon
-----------------------------------------------------------------------------

local L = AceLibrary("AceLocale-2.2"):new("RapidCast")

RapidCast = AceLibrary("AceAddon-2.0"):new("AceConsole-2.0", "AceEvent-2.0", "AceDB-2.0")

RapidCast:RegisterDB("RapidCastDB")
RapidCast:RegisterDefaults('profile', 
	{
		castTimePadding = 0.1,
		incorporateGCD = true,
	}
)

local consoleOptions = {
	type='group',
	args = {
		padding = {
			type = 'range',
			name = L["Cast Time Padding"],
			desc = L["Adjusts the time padding used as a safety margin for casting"],
			get = function() return RapidCast.db.profile.castTimePadding end,
			set = function(v) RapidCast.db.profile.castTimePadding = v end,
			min = 0.0,
			max = 1.0,
			step = 0.01,
		},
		incorporateGCD = {
			type = 'toggle',
			name = L["Incorporate Global Cooldown"],
			desc = L["Adjusts the casting bar to a minimum of 1.5 seconds from when it was cast"],
			get = function() return RapidCast.db.profile.incorporateGCD end,
			set = function(v) RapidCast.db.profile.incorporateGCD = v end,
		},
	},
}

RapidCast:RegisterChatCommand({"/rapidcast"}, consoleOptions)

function RapidCast:OnInitialize()
	self.triggerTime = nil
end

function RapidCast:OnEnable()
	self:RegisterEvent("UNIT_SPELLCAST_SENT", "SpellcastStartTrigger")
	self:RegisterEvent("UNIT_SPELLCAST_START", "SpellcastStartResponse")
	self:RegisterEvent("UNIT_SPELLCAST_CHANNEL_START", "SpellcastChannelStartResponse")
	self:RegisterEvent("UNIT_SPELLCAST_DELAYED", "SpellcastDelayed")
	self:RegisterEvent("UNIT_SPELLCAST_STOP", "SpellcastStop")
end

function RapidCast:OnDisable()
end

function RapidCast:SpellcastStartTrigger()
	self.triggerTime = GetTime()
	self.lag = 0
end

function RapidCast:SpellcastStop()
	self.lag = 0
end

function RapidCast:UpdateCastingBars(startTime, endTime, now)
	local duration = (endTime - startTime)*0.001

	if self.db.profile.incorporateGCD then
		if self.triggerTime and endTime*0.001 - self.triggerTime - self.lag < 1.5 then
			duration = 1.5
		end
	end

	startTime = startTime*0.001 - self.lag
	if (startTime > now) then
		startTime = now;
	end

	endTime = startTime + duration;
	if (now > endTime) then
		now = endTime;
	end

	self:UpdateCastingBarFrame(startTime, endTime, now)
end

function RapidCast:StartSpellCommon(startTime, endTime)
	local now = GetTime()

	if now > endTime*0.001 then 
		self.triggerTime = nil
		return 
	end
	
	self.lag = now - self.triggerTime - self.db.profile.castTimePadding
	if self.lag < 0 then self.lag = 0 end

	self:UpdateCastingBars(startTime, endTime, now)

	self.triggerTime = nil
end

function RapidCast:SpellcastStartResponse()
	if arg1 ~= "player" then return end

	self:SpellcastStart()

	if not self.triggerTime then return end

	local _, _, _, _, startTime, endTime = UnitCastingInfo("player");
	self:StartSpellCommon(startTime, endTime)
end

function RapidCast:SpellcastDelayed()
	if arg1 ~= "player" then return end

	self:SpellcastDelayed()

	local now = GetTime()
	local _, _, _, _, startTime, endTime = UnitCastingInfo("player")

	self:UpdateCastingBars(startTime, endTime, now)
end


function RapidCast:SpellcastChannelStartResponse()
	if arg1 ~= "player" then return end

	self:SpellcastChannelStart()

	if not self.triggerTime then return end

	local _, _, _, _, startTime, endTime = UnitChannelInfo("player");
	self:StartSpellCommon(startTime, endTime)
end

-----------------------------------------------------------------------------

-- TODO: Add support for other casting bars, etc. etc.
--       Maybe add support for multiple casting bars?

if oCB ~= nil then
	function RapidCast:UpdateCastingBarFrame(startTime, endTime, now)
		oCB.startTime = startTime
		oCB.endTime = endTime
		
		if oCB.channeling then
			oCB.maxValue = startTime
		else
			oCB.maxValue = endTime
		end
	
		oCB.frames.CastingBar.Bar:SetMinMaxValues(startTime, endTime)
		oCB.frames.CastingBar.Bar:SetValue(now)
	end

	function RapidCast:SpellcastStart()
	end

	function RapidCast:SpellcastChannelStart()
	end

	function RapidCast:SpellcastDelayed()
	end

elseif CBRipoff ~= nil then

	function RapidCast:UpdateCastingBarFrame(startTime, endTime, now)
		CBRipoff.startTime = startTime
		CBRipoff.endTime = endTime
		
		if CBRipoff.channeling then
			CBRipoff.maxValue = startTime
		else
			CBRipoff.maxValue = endTime
		end
	
		CBRipoff.frames.CastingBar.Bar:SetMinMaxValues(startTime, endTime)
		CBRipoff.frames.CastingBar.Bar:SetValue(now)
	end

	function RapidCast:SpellcastStart()
		-- We do this to ensure correct event ordering, which doesn't seem to occur with just dependencies
		-- ie. Call the CBRipoff's spellcast start, then call RapidCasts
		if CBRipoff:IsEventRegistered("UNIT_SPELLCAST_START") then
			CBRipoff:UnregisterEvent("UNIT_SPELLCAST_START")
		end
		CBRipoff:CastStart("player")
	end

	function RapidCast:SpellcastChannelStart()
		-- See above function. Same applies
		if CBRipoff:IsEventRegistered("UNIT_SPELLCAST_CHANNEL_START") then
			CBRipoff:UnregisterEvent("UNIT_SPELLCAST_CHANNEL_START")
		end
		CBRipoff:ChannelStart("player")
	end

	function RapidCast:SpellcastDelayed()
	end

elseif eCastingBar ~= nil then
	function RapidCast:UpdateCastingBarFrame(startTime, endTime, now)
		eCastingBar.startTime = startTime;
		eCastingBar.maxValue = endTime;
		eCastingBarStatusBar:SetMinMaxValues(startTime, endTime)
	end

	function RapidCast:SpellcastStart()
	end

	function RapidCast:SpellcastChannelStart()
	end

	function RapidCast:SpellcastDelayed()
	end

else
	function RapidCast:UpdateCastingBarFrame(startTime, endTime, now)
		CastingBarFrame.startTime = startTime;
		CastingBarFrame.endTime = endTime;
		CastingBarFrame.maxValue = endTime;
		CastingBarFrame:SetMinMaxValues(startTime, endTime);
		CastingBarFrame:SetValue(now);
	end

	function RapidCast:SpellcastStart()
	end

	function RapidCast:SpellcastChannelStart()
	end

	function RapidCast:SpellcastDelayed()
	end

end

if aUF and aUF.db.profile.units.player.CastBar then
	-- The difficulty with aUF is that the events can be registered at any time after
	-- startup, or even in between event triggers if the user selects various interface options
	--

	function RapidCast:GetPlayerAUFUnit()
		for obj,used in pairs(aUF.unitid.player) do
			if used then 
				return obj 
			end
		end
		return nil
	end

	RapidCast.aUFUpdateCastingBarFrame = RapidCast.UpdateCastingBarFrame
	RapidCast.aUFSpellcastStart = RapidCast.SpellcastStart 
	RapidCast.aUFSpellcastChannelStart = RapidCast.SpellcastChannelStart 
	RapidCast.aUFSpellcastDelayed = RapidCast.SpellcastDelayed

	function RapidCast:UpdateCastingBarFrame(startTime, endTime, now)
		local aUFPlayerUnit = self:GetPlayerAUFUnit()

		aUFPlayerUnit.spellCastData.startTime = startTime;
		aUFPlayerUnit.spellCastData.endTime = endTime;

		aUFPlayerUnit.CastBar:SetMinMaxValues(startTime, endTime);
		if (aUFPlayerUnit.spellCastData.casting) then
			aUFPlayerUnit.CastBar:SetValue(now);
		elseif (aUFPlayerUnit.spellCastData.channeling) then
			aUFPlayerUnit.CastBar:SetValue(endTime - (now-startTime));
		end
		
		self:aUFUpdateCastingBarFrame(startTime, endTime, now)
	end

	function RapidCast:SpellcastStart() 
		local aUFPlayerUnit = self:GetPlayerAUFUnit()
		aUFPlayerUnit:ObjectUnregisterEvent("UNIT_SPELLCAST_START") 
		aUFPlayerUnit:SpellcastStart() 

		self:aUFSpellcastStart()
	end

	function RapidCast:SpellcastChannelStart() 
		local aUFPlayerUnit = self:GetPlayerAUFUnit()
		aUFPlayerUnit:ObjectUnregisterEvent("UNIT_SPELLCAST_CHANNEL_START") 
		aUFPlayerUnit:SpellcastChannelStart() 

		self:aUFSpellcastChannelStart()
	end

	function RapidCast:SpellcastDelayed()
		local aUFPlayerUnit = self:GetPlayerAUFUnit()
		aUFPlayerUnit:ObjectUnregisterEvent("UNIT_SPELLCAST_DELAYED") 
		aUFPlayerUnit:SpellcastDelayed() 
		aUFPlayerUnit.spellCastData.delay = aUFPlayerUnit.spellCastData.delay - self.lag

		self:aUFSpellcastDelayed()
	end
end

-----------------------------------------------------------------------------
